from flask import Blueprint

from bolinette.routing import Defaults, Route, AccessToken


class Namespace:
    namespaces = []

    def __init__(self, service, url):
        self.service = service
        self.model = service.name
        self.url = url
        self.blueprint = Blueprint(self.model, __name__, url_prefix='/api' + url)
        self.route = NamespaceRoute(self)

    @staticmethod
    def init_namespaces(app):
        for namespace in Namespace.namespaces:
            app.register_blueprint(namespace)

    def register(self):
        Namespace.namespaces.append(self.blueprint)

    @property
    def defaults(self):
        return Defaults(self)


class NamespaceRoute:
    def __init__(self, namespace):
        self.namespace = namespace

    def __call__(self, rule, **options):
        def inner(func):
            endpoint = options.get('endpoint', func.__name__)
            methods = options.get('methods', ['GET'])
            roles = options.get('roles', [])
            access = options.get('access', (AccessToken.Required if len(roles)
                                            else AccessToken.Optional))
            expects = options.get('expects', None)
            returns = options.get('returns', None)
            route_rules = Route(func, self.namespace.url, rule, endpoint, methods,
                                access, expects, returns, roles)
            self.namespace.blueprint.add_url_rule(
                rule, endpoint, route_rules.process, methods=methods)
            return func

        return inner

    def expects(self, model, key='default', *, patch=False):
        return {
            'model': model,
            'key': key,
            'patch': patch
        }

    def returns(self, model, key='default', *, as_list=False, skip_none=False):
        return {
            'model': model,
            'key': key,
            'as_list': as_list,
            'skip_none': skip_none
        }



from bolinette.routing import Namespace
from bolinette.services import role_service

ns = Namespace(role_service, '/role')

ns.defaults.get_all(roles=['admin'])
ns.defaults.get_first_by('name', 'complete', roles=['admin'])

ns.register()



from flask import Flask
from flask_cors import CORS
from bolinette_cli import paths

from bolinette import env
from bolinette.database import init_db
from bolinette.jwt import init_jwt
from bolinette.routes import init_routes
from bolinette.documentation import init_docs
from bolinette.errors import init_error_handlers
from bolinette.routing import Namespace
from bolinette.mail import sender


class Bolinette:
    def __init__(self, name, **options):
        profile = options.get('profile')
        env_overrides = options.get('env', {})
        self.app = Flask(name, static_url_path='')
        CORS(self.app, supports_credentials=True)
        env.init_app(self, profile=profile, overrides=env_overrides)
        init_jwt(self.app)
        init_db(self)
        init_routes(self.app)
        init_docs(self.app)
        init_error_handlers(self.app)
        Namespace.init_namespaces(self.app)
        sender.init_app()



import flask

from bolinette import serialize


class Response:
    def build_message(self, code, status, messages=None, data=None):
        if messages is None:
            messages = []
        if not isinstance(messages, list):
            messages = [messages]
        body = {
            'code': code,
            'status': status,
            'messages': messages
        }
        if data is not None:
            body['data'] = data
        return body, code

    def ok(self, messages=None, data=None):
        return self.build_message(200, 'OK', messages, data)

    def created(self, messages=None, data=None):
        return self.build_message(201, 'CREATED', messages, data)

    def bad_request(self, messages=None, data=None):
        return self.build_message(400, 'BAD REQUEST', messages, data)

    def unauthorized(self, messages=None, data=None):
        return self.build_message(401, 'UNAUTHORIZED', messages, data)

    def forbidden(self, messages=None, data=None):
        return self.build_message(403, 'FORBIDDEN', messages, data)

    def not_found(self, messages=None, data=None):
        return self.build_message(404, 'NOT FOUND', messages, data)

    def method_not_allowed(self, messages=None, data=None):
        return self.build_message(405, 'METHOD NOT ALLOWED', messages, data)

    def conflict(self, messages=None, data=None):
        return self.build_message(409, 'CONFLICT', messages, data)

    def internal_server_error(self, messages=None, data=None):
        return self.build_message(500, 'INTERNAL SERVER ERROR', messages, data)

    def abort(self, message, code):
        res, mime = serialize(message)
        flask.abort(flask.Response(res, code, content_type=mime))


response = Response()



from bolinette.exceptions import ParamConflictError


class Mapper:
    def update(self, model, entity, params, patch=False):
        errors = []
        for column in model.__table__.columns:
            key = column.key
            if column.primary_key or (key not in params and patch):
                continue
            original = getattr(entity, key)
            new = params.get(key, None)
            if original == new:
                continue
            if column.unique and new is not None:
                if model.query.filter(getattr(model, key) == new).first() is not None:
                    errors.append((key, new))
            setattr(entity, key, new)
        if len(errors) > 0:
            raise ParamConflictError(params=errors)


mapper = Mapper()



import json

from flask import Blueprint, Response

from bolinette import marshalling


class Documentation:
    def __init__(self):
        self.app = None
        self.routes = []

    def add_route(self, route):
        self.routes.append(route)

    def process_definition(self, key, get_func):
        if isinstance(key, str):
            definition = get_func(key)
        else:
            definition = get_func(f'{key["model"]}.{key["key"]}')
        ret = {}
        for field in definition.fields:
            if isinstance(field, marshalling.Field):
                ret[field.name] = str(field.type)
            elif isinstance(field, marshalling.Definition):
                ret[field.name] = self.process_definition(field.key, get_func)
            elif isinstance(field, marshalling.List):
                ret[field.name] = [self.process_definition(field.element.key, get_func)]
        return ret

    def process_payload(self, expects):
        return self.process_definition(expects, marshalling.get_payload)

    def process_response(self, returns):
        return self.process_definition(returns, marshalling.get_response)

    def process_docs(self):
        ret = {}
        for route in self.routes:
            doc = route.func.__doc__
            if isinstance(doc, str):
                doc = doc.strip()
            if route.base_url not in ret:
                ret[route.base_url] = {}
            if route.url not in ret[route.base_url]:
                ret[route.base_url][route.url] = {}
            expects = route.expects
            if expects is not None:
                expects = self.process_payload(expects)
            returns = route.returns
            if returns is not None:
                returns = self.process_response(returns)
            for method in route.methods:
                ret[route.base_url][route.url][method.lower()] = {
                    'doc': doc,
                    'expects': expects,
                    'returns': returns
                }
        return ret


docs = Documentation()


def init_docs(app):
    def inner():
        return Response(json.dumps(docs.process_docs()), 200, mimetype='application/json')
    docs.app = app
    blueprint = Blueprint('bolinette_docs', __name__, url_prefix='/docs')
    blueprint.add_url_rule('', 'bolinette_docs', inner, methods=['GET'])
    app.register_blueprint(blueprint)



import random
import string

from flask import after_this_request
from flask_jwt_extended import (
    create_access_token, create_refresh_token, set_access_cookies, set_refresh_cookies,
    get_jwt_identity, unset_jwt_cookies
)
from bolinette import env, response, mail
from bolinette.exceptions import EntityNotFoundError, BadRequestError
from bolinette.routing import Namespace, current_user, AccessToken
from bolinette.services import user_service, role_service

ns = Namespace(user_service, '/user')


def _set_login_cookies(access_token, refresh_token):
    def inner(resp):
        set_access_cookies(resp, access_token)
        set_refresh_cookies(resp, refresh_token)
        return resp
    return inner


def _reset_cookies(access_token):
    def inner(resp):
        set_access_cookies(resp, access_token)
        return resp
    return inner


def _unset_cookies(resp):
    unset_jwt_cookies(resp)
    return resp


@ns.route('/me',
          methods=['GET'],
          access=AccessToken.Fresh,
          returns=ns.route.returns('user', 'private'))
def me():
    return response.ok('OK', current_user())


@ns.route('/me',
          methods=['PATCH'],
          access=AccessToken.Fresh,
          returns=ns.route.returns('user', 'private'),
          expects=ns.route.expects('user', 'register', patch=True))
def update_user(payload):
    user = user_service.patch(current_user(), payload)
    access_token = create_access_token(identity=user.username, fresh=True)
    refresh_token = create_refresh_token(identity=user.username)
    after_this_request(_set_login_cookies(access_token, refresh_token))
    return response.ok('user.updated', user)


@ns.route('/info',
          methods=['GET'],
          access=AccessToken.Required,
          returns=ns.route.returns('user', 'private'))
def info():
    return response.ok('OK', current_user())


@ns.route('/login',
          methods=['POST'],
          returns=ns.route.returns('user', 'private'),
          expects=ns.route.expects('user', 'login'))
def login(payload):
    """
    Logs the user in with the provided credentials
    """
    username = payload['username']
    password = payload['password']
    try:
        user = user_service.get_by_username(username)
    except EntityNotFoundError:
        return response.unauthorized('user.login.wrong_credentials')
    if user is not None:
        if user_service.check_password(user, password):
            access_token = create_access_token(identity=user.username, fresh=True)
            refresh_token = create_refresh_token(identity=user.username)
            after_this_request(_set_login_cookies(access_token, refresh_token))
            return response.ok('user.login.success', user)
    return response.unauthorized('user.login.wrong_credentials')


@ns.route('/logout',
          methods=['POST'])
def logout():
    after_this_request(_unset_cookies)
    return response.ok('user.logout.success')


@ns.route('/register',
          methods=['POST'],
          returns=ns.route.returns('user', 'private'),
          expects=ns.route.expects('user', 'register'))
def register(payload):
    if env.init.get('ADMIN_REGISTER_ONLY', True):
        raise BadRequestError('global.register.admin_only')
    if current_user() is not None:
        raise BadRequestError('global.register.logged_in')
    user = user_service.create(payload)
    access_token = create_access_token(identity=user.username, fresh=True)
    refresh_token = create_refresh_token(identity=user.username)
    after_this_request(_set_login_cookies(access_token, refresh_token))
    return response.created('user.registered', user)


@ns.route('/register/admin',
          methods=['POST'],
          roles=['adminn'],
          returns=ns.route.returns('user', 'private'),
          expects=ns.route.expects('user', 'admin_register'))
def admin_register(payload):
    send_mail = payload.pop('send_mail')
    payload['password'] = ''.join(random.choices(string.ascii_lowercase, k=32))
    user = user_service.create(payload)
    if send_mail:
        mail.sender.send(payload['email'], 'Welcome!', 'Welcome to Bolinette!')
    return response.created('user.registered', user)


@ns.route('/token/refresh',
          methods=['POST'],
          access=AccessToken.Refresh)
def refresh():
    identity = get_jwt_identity()
    access_token = create_access_token(identity=identity, fresh=False)
    after_this_request(_reset_cookies(access_token))
    return response.ok('user.token.refreshed')


ns.defaults.get_all('private', roles=['admin'])

ns.defaults.get_first_by('username', returns='private', roles=['admin'])


@ns.route('/<username>/roles',
          methods=['POST'],
          roles=['admin'],
          expects=ns.route.expects('role'),
          returns=ns.route.returns('user', 'private'))
def add_user_role(username, payload):
    user = user_service.get_by_username(username)
    role = role_service.get_by_name(payload['name'])
    user_service.add_role(user, role)
    return response.created(f'user.roles.added:{user.username}:{role.name}', user)


@ns.route('/<username>/roles/<role>',
          methods=['DELETE'],
          roles=['admin'],
          returns=ns.route.returns('user', 'private'))
def delete_user_role(username, role):
    user = user_service.get_by_username(username)
    role = role_service.get_by_name(role)
    user_service.remove_role(current_user(), user, role)
    return response.ok(f'user.roles.removed:{user.username}:{role.name}', user)


ns.register()



import inspect

from flask import request, Response
from flask_sqlalchemy import Pagination

from bolinette import transaction, marshalling, validate, docs, serialize
from bolinette.routing import AccessToken


class Route:
    @staticmethod
    def _func_accepts_keywords(func):
        return any(param for param in inspect.signature(func).parameters.values() if param.kind == param.VAR_KEYWORD)

    def __init__(self, func, base_url, url, endpoint, methods, access, expects, returns, roles):
        self.func = func
        self.base_url = base_url
        self.url = url
        self.endpoint = endpoint
        self.methods = methods
        self.access = access
        self.expects = expects
        self.returns = returns
        self.roles = roles
        docs.add_route(self)

    def process(self, *args, **kwargs):
        self.access.check(self.roles)
        payload = request.get_json(silent=True) or {}
        if len(request.args) and Route._func_accepts_keywords(self.func):
            kwargs['args'] = dict(request.args)
        with transaction:
            if self.expects is not None:
                def_key = f'{self.expects["model"]}.{self.expects["key"]}'
                exp_def = marshalling.get_payload(def_key)
                kwargs['payload'] = validate.payload(
                    exp_def, payload, self.expects.get('patch', False))
                marshalling.link_foreign_entities(exp_def, kwargs['payload'])

            res, code = self.func(*args, **kwargs)

        if res.get('data') is not None and isinstance(res['data'], Pagination):
            res['pagination'] = {
                'page': res['data'].page,
                'per_page': res['data'].per_page,
                'total': res['data'].total,
            }
            res['data'] = res['data'].items

        if self.returns is not None:
            def_key = f'{self.returns["model"]}.{self.returns["key"]}'
            ret_def = marshalling.get_response(def_key)
            if res.get('data') is not None:
                res['data'] = marshalling.marshall(
                    ret_def, res['data'], self.returns.get('skip_none', False),
                    self.returns.get('as_list', False))

        res, mime = serialize(res)
        return Response(res, code, mimetype=mime)



from bolinette import response


class Defaults:
    def __init__(self, namespace):
        self.model = namespace.model
        self.service = namespace.service
        self.route = namespace.route

    def _process_access_options(self, options):
        access_options = {}
        if 'access' in options:
            access_options['access'] = options['access']
        if 'roles' in options:
            access_options['roles'] = options['roles']
        return access_options

    def get_all(self, returns='default', **options):
        def route(**params):
            pagination = None
            order_by = []
            if 'args' in params:
                if 'page' in params['args'] or 'per_page' in params['args']:
                    pagination = {
                        'page': int(params['args'].get('page', 1)),
                        'per_page': int(params['args'].get('per_page', 20)),
                        'error_out': False
                    }
                if 'order_by' in params['args']:
                    columns = params['args']['order_by'].split(',')
                    for column in columns:
                        order_args = column.split(':')
                        col_name = order_args[0]
                        order_way = order_args[1] if len(order_args) > 1 else 'asc'
                        order_by.append((col_name, order_way == 'asc'))
            return response.ok('OK', self.service.get_all(pagination, order_by))

        access_options = self._process_access_options(options)
        self.route('',
                   methods=['GET'],
                   endpoint=f'get_all_{self.model}',
                   returns=self.route.returns(self.model, returns, as_list=True),
                   **access_options)(route)

    def get_one(self, returns='default', **options):
        def route(**params):
            m_id = params.get('id')
            return response.ok('OK', self.service.get(m_id))

        access_options = self._process_access_options(options)
        self.route('/<id>',
                   methods=['GET'],
                   endpoint=f'get_one_{self.model}',
                   returns=self.route.returns(self.model, returns),
                   **access_options)(route)

    def get_first_by(self, key, returns='default', **options):
        def route(**params):
            m_id = params.get('id')
            return response.ok('OK', self.service.get_first_by(key, m_id))

        access_options = self._process_access_options(options)
        self.route('/<id>',
                   methods=['GET'],
                   endpoint=f'get_one_{self.model}',
                   returns=self.route.returns(self.model, returns),
                   **access_options)(route)

    def create(self, returns='default', expects='default', **options):
        def route(**params):
            payload = params.get('payload')
            return response.created(f'{self.model}.created', self.service.create(payload))

        access_options = self._process_access_options(options)
        self.route('',
                   methods=['POST'],
                   endpoint=f'create_{self.model}',
                   returns=self.route.returns(self.model, returns),
                   expects=self.route.expects(self.model, expects),
                   **access_options)(route)

    def update(self, returns='default', expects='default', **options):
        def route(**params):
            m_id = params.get('id')
            payload = params.get('payload')
            entity = self.service.get(m_id)
            return response.ok(f'{self.model}.updated', self.service.update(entity, payload))

        access_options = self._process_access_options(options)
        self.route('/<id>',
                   methods=['PUT'],
                   endpoint=f'update_{self.model}',
                   returns=self.route.returns(self.model, returns),
                   expects=self.route.expects(self.model, expects),
                   **access_options)(route)

    def patch(self, returns='default', expects='default', **options):
        def route(**params):
            m_id = params.get('id')
            payload = params.get('payload')
            entity = self.service.get(m_id)
            return response.ok(f'{self.model}.updated', self.service.patch(entity, payload))

        access_options = self._process_access_options(options)
        self.route('/<id>',
                   methods=['PATCH'],
                   endpoint=f'patch_{self.model}',
                   returns=self.route.returns(self.model, returns),
                   expects=self.route.expects(self.model, expects, patch=True),
                   **access_options)(route)

    def delete(self, returns='default', **options):
        def route(**params):
            m_id = params.get('id')
            entity = self.service.get(m_id)
            return response.ok(f'{self.model}.deleted', self.service.delete(entity))

        access_options = self._process_access_options(options)
        self.route('/<id>',
                   methods=['DELETE'],
                   endpoint=f'delete_{self.model}',
                   returns=self.route.returns(self.model, returns),
                   **access_options)(route)



from functools import wraps
import pytest

from bolinette import Bolinette, db
from bolinette.testing import TestClient


@pytest.fixture
def client():
    return TestClient(Bolinette(__name__, profile='test',
                                env={'DBMS': 'SQLITE', 'SECRET_KEY': 'super secret'}))


def bolitest(**options):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            db.drop_all()
            db.create_all()
            if 'before' in options and callable(options['before']):
                options['before']()
            db.session.commit()
            func(*args, **kwargs)
            if 'after' in options and callable(options['after']):
                options['after']()
            db.drop_all()
        return inner
    return wrapper



from bolinette import bcrypt
from bolinette.models import User
from bolinette.services import BaseService
from bolinette.exceptions import ForbiddenError, BadRequestError


class UserService(BaseService):
    @staticmethod
    def encrypt_password(password):
        return bcrypt.generate_password_hash(password).decode('utf-8')

    def __init__(self):
        super().__init__(User, 'user')

    def get_by_username(self, username):
        return self.get_first_by('username', username)

    def get_by_email(self, email):
        return self.get_first_by('email', email)

    def check_password(self, user, password):
        return bcrypt.check_password_hash(user.password, password)

    def create(self, params):
        if params.get('password'):
            params['password'] = UserService.encrypt_password(params['password'])
        return super().create(params)

    def update(self, entity, params):
        if params.get('password'):
            params['password'] = UserService.encrypt_password(params['password'])
        return super().update(entity, params)

    def patch(self, entity, params):
        if params.get('password'):
            params['password'] = UserService.encrypt_password(params['password'])
        return super().patch(entity, params)

    def add_role(self, user, role):
        if role.name == 'root':
            raise ForbiddenError('role.root.forbidden')
        if role in user.roles:
            raise BadRequestError(f'user.roles.exists:{user.username}:{role.name}')
        user.roles.append(role)

    def remove_role(self, current_user, user, role):
        if role.name == 'root':
            raise ForbiddenError('role.root.forbidden')
        if (current_user.username == user.username
                and role.name == 'admin'
                and not current_user.has_role('root')):
            raise ForbiddenError('role.admin.no_self_demotion')
        if role not in user.roles:
            raise BadRequestError(f'user.roles.not_found:{user.username}:{role.name}')
        user.roles.remove(role)


user_service = UserService()



from enum import Enum, unique, auto

from flask_jwt_extended import (
    verify_jwt_in_request, verify_jwt_in_request_optional, verify_fresh_jwt_in_request,
    verify_jwt_refresh_token_in_request, current_user
)

from bolinette import response, serialize


@unique
class AccessToken(Enum):
    All = auto()
    Optional = auto()
    Required = auto()
    Fresh = auto()
    Refresh = auto()

    def check_roles(self, roles):
        user_roles = set(map(lambda r: r.name, current_user.roles))
        if 'root' not in user_roles and not len(user_roles.intersection(set(roles))):
            response.abort(*response.forbidden(f'user.forbidden:{",".join(roles)}'))

    def check(self, roles):
        _functions[self.value]()
        if len(roles):
            self.check_roles(roles)


_functions = {
    AccessToken.All.value: lambda: None,
    AccessToken.Optional.value: verify_jwt_in_request_optional,
    AccessToken.Required.value: verify_jwt_in_request,
    AccessToken.Fresh.value: verify_fresh_jwt_in_request,
    AccessToken.Refresh.value: verify_jwt_refresh_token_in_request
}



class MarshallingObject:
    pass


class Field(MarshallingObject):
    def __init__(self, field_type, name, *, default=None, required=False,
                 function=None, formatting=None):
        self.type = field_type
        self.name = name
        self.required = required
        self.default = default
        self.function = function
        self.formatting = formatting

    def __repr__(self):
        return f'<MarshallingField {self.name}:{self.type}>'


class List(MarshallingObject):
    def __init__(self, name, element):
        self.name = name
        self.element = element

    def __repr__(self):
        return f'<MarshallingList {self.name}:[{repr(self.element)}]>'


class Definition(MarshallingObject):
    def __init__(self, name, model, key='default'):
        self.fields = []
        self.name = name
        self.key = f'{model}.{key}'

    def __repr__(self):
        return f'<MarshallingModel {self.key}>'



from bolinette import env, seeder, transaction
from bolinette.services import role_service, user_service


@seeder
def role_seeder():
    with transaction:
        role_service.create({'name': 'root'})
        role_service.create({'name': 'admin'})


@seeder
def dev_user_seeder():
    if env['PROFILE'] == 'development':
        with transaction:
            root = role_service.get_by_name('root')
            admin = role_service.get_by_name('admin')
            root_usr = user_service.create({
                'username': 'root',
                'password': 'root',
                'email': f'root@localhost'
            })
            root_usr.roles.append(root)
            root_usr.roles.append(admin)

            dev0 = role_service.create({'name': 'dev0'})
            dev1 = role_service.create({'name': 'dev1'})
            dev2 = role_service.create({'name': 'dev2'})
            roles = [dev0, dev1, dev2]

            for i in range(10):
                user = user_service.create({
                    'username': f'user_{i}',
                    'password': 'test',
                    'email': f'user{i}@test.com'
                })
                user.roles.append(roles[i % 3])
                user.roles.append(roles[(i + 1) % 3])



import json
import dicttoxml
from htmlmin import minify
from flask import request

from bolinette_cli import templating

from bolinette import env


class Serializer:
    def __init__(self, mime, priority):
        self.mime = mime
        self.priority = priority

    def serialize(self, response):
        pass


class HTMLSerializer(Serializer):
    def __init__(self):
        super().__init__('text/html', 0)

    def serialize(self, response):
        return minify(templating.render(
            env.internal_path('files', 'default.html.jinja2'), {'response': response})
        )


class JSONSerializer(Serializer):
    def __init__(self):
        super().__init__('application/json', 1)

    def serialize(self, response):
        return json.dumps(response)


class XMLSerializer(Serializer):
    def __init__(self):
        super().__init__('application/xml', 2)

    def serialize(self, response):
        return dicttoxml.dicttoxml(response)


class Serializers:
    def __init__(self, functions=None):
        self.priorities = []
        self.serializers = {}
        if functions is not None:
            for serializer in functions:
                self.add(serializer)

    def add(self, serializer):
        self.serializers[serializer.mime] = serializer
        self.priorities = sorted(self.priorities + [serializer], key=lambda s: s.priority)

    def get(self, mime):
        return self.serializers.get(mime, None)

    @property
    def default(self):
        return self.priorities[0]


serializers = Serializers([
    HTMLSerializer(),
    JSONSerializer(),
    XMLSerializer(),
])


def serialize(response):
    mime = request.headers.get('Accept', 'application/json')
    serializer = serializers.get(mime) or serializers.default
    return serializer.serialize(response), serializer.mime



from functools import wraps

from sqlalchemy.exc import SQLAlchemyError
from bolinette_cli import logger

from bolinette import db, response, serialize
from bolinette.exceptions import APIError


class Transaction:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            db.session.rollback()
            res, code = None, None
            if issubclass(exc_type, APIError):
                res, code = exc_val.response
            else:
                logger.error(str(exc_val))
                res, code = response.internal_server_error(str(exc_val))
            if res is not None and code is not None:
                response.abort(res, code)
        else:
            try:
                db.session.commit()
            except SQLAlchemyError as err:
                db.session.rollback()
                res, code = response.internal_server_error(f'global.internal_error:{err}')
                response.abort(res, code)


transaction = Transaction()



import json

from bolinette import db


class TestClient:
    def __init__(self, bolinette):
        self.bolinette = bolinette
        self.bolinette.app.app_context().push()
        self.client = self.bolinette.app.test_client()
        self.cookies = {}

    def payload(self, **payload):
        if 'csrf_access_token' in self.cookies:
            payload['headers'] = {}
            payload['headers']['X-CSRF-TOKEN'] = self.cookies['csrf_access_token']
        return payload

    def parse_cookies(self, headers):
        for key, header in headers:
            if key == 'Set-Cookie':
                split = header.split(';')[0].split('=')
                self.cookies[split[0]] = split[1]

    def set_up(self):
        db.drop_all()
        db.create_all()

    def tear_down(self):
        self.client.cookie_jar.clear()
        db.drop_all()

    def post(self, path, data=None):
        if data is None:
            data = {}
        payload = self.payload(data=json.dumps(data),
                               content_type='application/json')
        res = self.client.post(f'/api{path}', **payload)
        self.parse_cookies(res.headers)
        return json.loads(res.data)

    def put(self, path, data=None):
        if data is None:
            data = {}
        payload = self.payload(data=json.dumps(data),
                               content_type='application/json')
        res = self.client.put(f'/api{path}', **payload)
        self.parse_cookies(res.headers)
        return json.loads(res.data)

    def patch(self, path, data=None):
        if data is None:
            data = {}
        payload = self.payload(data=json.dumps(data),
                               content_type='application/json')
        res = self.client.patch(f'/api{path}', **payload)
        self.parse_cookies(res.headers)
        return json.loads(res.data)

    def get(self, path):
        res = self.client.get(f'/api{path}', **self.payload())
        return json.loads(res.data)

    def delete(self, path):
        return json.loads(self.client.delete(f'/api{path}', **self.payload()).data)



from flask import Response

from bolinette import response, env, serialize


def _error_handler(error, message, func):
    res, code = func(error.description if env['DEBUG'] else message)
    res, mime = serialize(res)
    return Response(res, code, mimetype=mime)


def init_error_handlers(app):
    app.errorhandler(404)(
        lambda error: _error_handler(error, 'global.resource_not_found', response.not_found)
    )
    app.errorhandler(500)(
        lambda error: _error_handler(error, 'global.internal_error', response.internal_server_error)
    )



from bolinette import db, marshalling


class Role(db.Model):
    __tablename__ = 'roles'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), unique=True, nullable=False)

    def __repr__(self):
        return f'<Role {self.name}>'

    @staticmethod
    def payloads():
        yield [
            marshalling.Field(marshalling.types.string, 'name')
        ]

    @staticmethod
    def responses():
        yield [
            marshalling.Field(marshalling.types.string, 'name')
        ]
        yield 'complete', [
            marshalling.Field(marshalling.types.string, 'name'),
            marshalling.List('users', marshalling.Definition('user', 'user'))
        ]


marshalling.register(Role, 'role')



from flask import send_from_directory

from bolinette import env


def index():
    return send_from_directory(env['WEBAPP_FOLDER'], 'index.html')


def init_routes(app):
    app.add_url_rule('/', 'index', view_func=index)



class MarshallingType:
    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__.lower()


class String(MarshallingType):
    pass


class Integer(MarshallingType):
    pass


class Float(MarshallingType):
    pass


class Boolean(MarshallingType):
    pass


class Date(MarshallingType):
    pass


class Email(MarshallingType):
    pass


class Password(MarshallingType):
    pass


class ForeignKey(MarshallingType):
    def __init__(self, model, key):
        self.model = model
        self.key = key


class TypesClasses:
    String = String
    Integer = Integer
    Float = Float
    Boolean = Boolean
    Date = Date
    Email = Email
    Password = Password
    ForeignKey = ForeignKey


class Types:
    classes = TypesClasses

    @property
    def string(self):
        return String()

    @property
    def integer(self):
        return Integer()

    @property
    def float(self):
        return Float()

    @property
    def boolean(self):
        return Boolean()

    @property
    def date(self):
        return Date()

    @property
    def email(self):
        return Email()

    @property
    def password(self):
        return Password()

    def foreign_key(self, model, key):
        return ForeignKey(model, key)


types = Types()



from bolinette.environment import env
from bolinette.bcrypt import bcrypt
from bolinette.database import db, seeder
from bolinette.serializing import serialize
from bolinette.response import response
from bolinette.validator import validate
from bolinette.documentation import docs
from bolinette.transaction import transaction
from bolinette.mapper import mapper
from bolinette.jwt import jwt
import bolinette.controllers
import bolinette.seeders
from bolinette.bolinette import Bolinette



import os
from functools import wraps

from flask_sqlalchemy import SQLAlchemy
from bolinette_cli import logger

from bolinette import env

_seeder_funcs = []

db = SQLAlchemy()


def create_db_uri():
    dbms = env.get('DBMS', 'SQLITE').lower()
    if dbms == 'sqlite':
        return 'sqlite:///' + env.instance_path(env.get('SQLITE_FILE', f'{env["PROFILE"]}.db'))
    if dbms == 'memory':
        return 'sqlite://'
    if dbms == 'postgresql':
        return 'postgresql://' + env['DB_URL']
    logger.error(f'Unknown database system "{dbms}"')
    exit(1)


def init_db(bolinette):
    bolinette.app.config['SQLALCHEMY_DATABASE_URI'] = create_db_uri()
    bolinette.app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True
    db.init_app(bolinette.app)


def seeder(func):
    _seeder_funcs.append(func)

    @wraps(func)
    def inner(*args, **kwargs):
        func(*args, **kwargs)

    return inner


def run_seeders():
    for func in _seeder_funcs:
        func()



from bolinette import db, marshalling, env

users_roles = db.Table('users_roles',
                       db.Column('u_id', db.Integer, db.ForeignKey('users.id'), primary_key=True),
                       db.Column('r_id', db.Integer, db.ForeignKey('roles.id'), primary_key=True))


class User(db.Model):
    __tablename__ = 'users'

    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(255), unique=True, nullable=False)
    password = db.Column(db.String(255), nullable=False)
    email = db.Column(db.String(255), unique=env.init['USER_EMAIL_REQUIRED'],
                      nullable=(not env.init['USER_EMAIL_REQUIRED']))

    roles = db.relationship('Role', secondary=users_roles, lazy='subquery',
                            backref=db.backref('users', lazy=True))

    def has_role(self, role):
        return any(filter(lambda r: r.name == role, self.roles))

    @staticmethod
    def payloads():
        yield 'register', [
            marshalling.Field(marshalling.types.string, 'username', required=True),
            marshalling.Field(marshalling.types.email, 'email', required=True),
            marshalling.Field(marshalling.types.password, 'password', required=True)
        ]
        yield 'admin_register', [
            marshalling.Field(marshalling.types.string, 'username', required=True),
            marshalling.Field(marshalling.types.email, 'email', required=True),
            marshalling.Field(marshalling.types.boolean, 'send_mail', required=True)
        ]
        yield 'login', [
            marshalling.Field(marshalling.types.string, 'username', required=True),
            marshalling.Field(marshalling.types.password, 'password', required=True)
        ]

    @staticmethod
    def responses():
        yield [
            marshalling.Field(marshalling.types.string, 'username')
        ]
        yield 'private', [
            marshalling.Field(marshalling.types.string, 'username'),
            marshalling.Field(marshalling.types.email, 'email'),
            marshalling.List('roles', marshalling.Definition('role', 'role'))
        ]

    def __repr__(self):
        return f'<User {self.username}>'


marshalling.register(User, 'user')



from flask_jwt_extended import get_jwt_identity

from bolinette.exceptions import APIError
from bolinette.services import user_service


def current_user():
    try:
        return user_service.get_by_username(get_jwt_identity())
    except APIError:
        return None



from bolinette.routing.access import AccessToken
from bolinette.routing.route import Route
from bolinette.routing.defaults import Defaults
from bolinette.routing.namespace import Namespace
from bolinette.routing.current_user import current_user
